package terminal;

import static terminal.ControlSequenceHandler.isEscape;
import static util.Characters.EOF;

import java.io.*;
import java.util.concurrent.*;

import com.googlecode.lanterna.*;
import com.googlecode.lanterna.input.*;
import com.googlecode.lanterna.terminal.*;

import terminal.SafePipedStream.*;

public class LispTerminal {

    public static final char END_OF_SEGMENT = 'x';

    private IOSafeTerminal terminal;
    private SafePipedOutputStream inputWriter;
    private SafePipedInputStream outputReader;
    private ExecutorService executorService;
    private ControlSequenceHandler controlSequenceHandler;
    private TerminalSize terminalSize;
    private int originColumn;
    private int originRow;
    private String inputLine;
    private String outputSegment;
    private boolean isFinished;

    public LispTerminal(IOSafeTerminal terminal, PipedOutputStream inputWriter, PipedInputStream outputReader) {
        // FIXME - add resize handler
        this.terminal = terminal;
        this.inputWriter = new SafePipedOutputStream(inputWriter);
        this.outputReader = new SafePipedInputStream(outputReader);
        this.executorService = Executors.newFixedThreadPool(2);
        this.controlSequenceHandler = new ControlSequenceHandler();
        this.terminalSize = terminal.getTerminalSize();
        this.originColumn = terminal.getCursorPosition().getColumn();
        this.originRow = terminal.getCursorPosition().getRow();
        this.inputLine = "";
        this.outputSegment = "";
        this.isFinished = false;

        addResizeListener();
    }

    private void addResizeListener() {
        terminal.addResizeListener(new TerminalResizeListener() {

            @Override
            public void onResized(Terminal terminal, TerminalSize newSize) {
                resize();
            }
        });
    }

    private void resize() {
        terminalSize = terminal.getTerminalSize();
    }

    public void run() {
        executorService.execute(this::readInput);
        executorService.execute(this::writeOutput);
        executorService.shutdown();
    }

    public void readInput() {
        while (!isFinished)
            processNextKey();
    }

    private void processNextKey() {
        KeyStroke keyStroke = getKeyStroke();

        if (keyStroke != null)
            handleKey(keyStroke);
        else
            takeNap();
    }

    private KeyStroke getKeyStroke() {
        KeyStroke keyStroke = null;

        try {
            keyStroke = terminal.pollInput();
        } catch (IllegalStateException e) {
            // Issue #299
            terminal.putCharacter('\n');
            terminal.close();
            System.exit(0);
        }

        return keyStroke;
    }

    private synchronized void handleKey(KeyStroke keyStroke) {
        doKey(keyStroke);
        terminal.flush();
    }

    private synchronized void doKey(KeyStroke keyStroke) {
        KeyType keyType = keyStroke.getKeyType();

        if (keyStroke.isCtrlDown())
            doControlKey(keyStroke);
        else if (keyType == KeyType.ArrowLeft)
            moveCursorLeft();
        else if (keyType == KeyType.ArrowRight)
            moveCursorRight();
        else if (keyType == KeyType.Enter)
            doEnter();
        else if (keyType == KeyType.Backspace)
            doBackspace();
        else if (keyType == KeyType.Delete)
            doDelete();
        else if (keyType == KeyType.Character)
            doCharacter(keyStroke);
    }

    private synchronized void doControlKey(KeyStroke keyStroke) {
        KeyType keyType = keyStroke.getKeyType();

        if (keyType == KeyType.Character)
            doControlCharacter(keyStroke);
    }

    private synchronized void doControlCharacter(KeyStroke keyStroke) {
        if (keyStroke.getCharacter() == 'd')
            doControlD();
    }

    private synchronized void doControlD() {
        doEnter();
        finish();
    }

    private synchronized void moveCursorLeft() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveLeft(cursorPosition))
            retractCursor(cursorPosition);
    }

    private synchronized boolean isPossibleToMoveLeft(TerminalPosition cursorPosition) {
        return getDistanceFromOrigin(cursorPosition) > 0;
    }

    private synchronized int getDistanceFromOrigin(TerminalPosition cursorPosition) {
        int columnDifference = cursorPosition.getColumn() - originColumn;
        int rowDifference = cursorPosition.getRow() - originRow;
        int totalColumns = terminalSize.getColumns();

        return columnDifference + (totalColumns * rowDifference);
    }

    private synchronized TerminalPosition getLeadingEdge() {
        int inputLength = inputLine.length();
        int totalColumns = terminalSize.getColumns();
        int rowDifference = inputLength / totalColumns;
        int columnDifference = inputLength % totalColumns;

        return new TerminalPosition(originColumn + columnDifference, originRow + rowDifference);
    }

    private synchronized void retractCursor(TerminalPosition cursorPosition) {
        TerminalPosition newPosition = cursorPosition.withRelativeColumn(-1);

        if (isAtStartOfRow(cursorPosition))
            newPosition = cursorPosition.withColumn(terminalSize.getColumns()).withRelativeRow(-1);

        terminal.setCursorPosition(newPosition);

    }

    private boolean isAtStartOfRow(TerminalPosition cursorPosition) {
        return cursorPosition.getColumn() == 0;
    }

    private synchronized void moveCursorRight() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition))
            advanceCursor(cursorPosition);
    }

    private synchronized boolean isPossibleToMoveRight(TerminalPosition cursorPosition) {
        return getDistanceFromOrigin(cursorPosition) < inputLine.length();
    }

    private synchronized void advanceCursor(TerminalPosition cursorPosition) {
        if (isAtEndOfRow(cursorPosition))
            moveCursorToNextRow(cursorPosition);
        else
            terminal.setCursorPosition(cursorPosition.withRelativeColumn(1));
    }

    private synchronized boolean isAtEndOfRow(TerminalPosition cursorPosition) {
        return cursorPosition.getColumn() >= terminalSize.getColumns() - 1;
    }

    private void moveCursorToNextRow(TerminalPosition cursorPosition) {
        if (isEndOfBuffer())
            createNewRowForCursor();
        else
            terminal.setCursorPosition(cursorPosition.withColumn(0).withRelativeRow(1));
    }

    private boolean isEndOfBuffer() {
        return terminal.getCursorPosition().getRow() == terminalSize.getRows() - 1;
    }

    private void createNewRowForCursor() {
        TerminalPosition originalPosition = terminal.getCursorPosition();
        terminal.putCharacter('\n');
        terminal.setCursorPosition(originalPosition.withColumn(0));
        --originRow;
    }

    private synchronized void doEnter() {
        moveCursorToEndOfInput();
        terminal.putCharacter('\n');
        inputLine += "\n";
        inputWriter.write(inputLine.getBytes());
        inputWriter.flush();
        inputLine = "";
        updateOrigin();
    }

    private synchronized void moveCursorToEndOfInput() {
        terminal.setCursorPosition(getLeadingEdge());
    }

    private void updateOrigin() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();
        originColumn = cursorPosition.getColumn();
        originRow = cursorPosition.getRow();
    }

    private synchronized void doBackspace() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveLeft(cursorPosition)) {
            String remaining = inputLine.substring(getDistanceFromOrigin(cursorPosition), inputLine.length());
            inputLine = inputLine.substring(0, getDistanceFromOrigin(cursorPosition) - 1) + remaining;

            retractCursor(cursorPosition);

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            terminal.putCharacter(' ');

            retractCursor(cursorPosition);
        }
    }

    private synchronized void doDelete() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition)) {
            String remaining = inputLine.substring(getDistanceFromOrigin(cursorPosition) + 1, inputLine.length());
            inputLine = inputLine.substring(0, getDistanceFromOrigin(cursorPosition)) + remaining;

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            terminal.putCharacter(' ');
            terminal.setCursorPosition(cursorPosition);
        }
    }

    private synchronized void doCharacter(KeyStroke keyStroke) {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition)) {
            String remaining = keyStroke.getCharacter()
                               + inputLine.substring(getDistanceFromOrigin(cursorPosition), inputLine.length());
            inputLine = inputLine.substring(0, getDistanceFromOrigin(cursorPosition)) + remaining;

            // FIXME - must have a way to push remainder on to a new line at the end of the buffer

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            advanceCursor(cursorPosition);
        } else {
            terminal.putCharacter(keyStroke.getCharacter());
            inputLine += keyStroke.getCharacter();

            advanceCursor(cursorPosition);
        }
    }

    private void takeNap() {
        try {
            Thread.sleep(1);
        } catch (InterruptedException e) {
            isFinished = true;
        }
    }

    public void writeOutput() {
        for (int c = outputReader.read(); c != EOF; c = outputReader.read())
            processOutput((char) c);

        terminal.setCursorVisible(true);
        terminal.flush();
        terminal.close();
    }

    private synchronized void processOutput(char c) {
        if (isEscape(c))
            parseControlSequence();
        else if (isEndOfSegment(c))
            printSegment();
        else
            outputSegment += c;
    }

    private synchronized void parseControlSequence() {}

    private synchronized boolean isEndOfSegment(char c) {
        return c == END_OF_SEGMENT;
    }

    private synchronized void printSegment() {
        terminal.setCursorVisible(false);
        moveCursorToEndOfInput();

        for (char c : outputSegment.toCharArray())
            terminal.putCharacter(c);

        moveCursorToNextRowIfNecessary();
        terminal.flush();
        terminal.setCursorVisible(true);
        outputSegment = "";
        updateOrigin();
    }

    private synchronized void moveCursorToNextRowIfNecessary() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isAtEndOfRow(cursorPosition))
            moveCursorToNextRow(cursorPosition);
    }

    public void finish() {
        isFinished = true;
        inputWriter.close();
    }

}
